#include "stdafx.h"
#include <bzsddk/bzsddk.h>
#include <bzscmn/string.h>
#include <string>
#include <vector>
#include <map>
#include <bzsddk/scsi/VirtualSCSIPDO.h>
#include "VirtualCDDevice.h"
#include "../ImageFormats/ImageFormats.h"
#include "DeviceControl.h"

using namespace DDK;


class BazisVirtualCDBus : public _GenericIndexedBus<VirtualSCSIPDO, String, _CaseInsensitiveLess<String>>
{
private:
	void CleanupOldLinksAndRegisterNewDeviceLink(const String &deviceId, char SuggestedDriveLetter)
	{
		static const wchar_t wszVCDPrefix[] = L"\\??\\BazisVirtualCDBus#StandardDevice#";
		RegistryKey key(L"\\Registry\\Machine\\System\\MountedDevices");
		TypedBuffer<KEY_VALUE_FULL_INFORMATION> info;

		for (unsigned i = 0; ; i++)
		{
			if (!key.EnumValue(i, &info))
				break;
			if ((info->DataLength / sizeof(wchar_t)) < __countof(wszVCDPrefix))
				continue;
			PWCHAR pwszData = (PWCHAR)(((char *)info.GetData()) + info->DataOffset);
			if (memcmp(pwszData, wszVCDPrefix, sizeof(wszVCDPrefix) - sizeof(wszVCDPrefix[0])))
				continue;
			if (key.DeleteValue(info->Name, info->NameLength / sizeof(wchar_t)))
				i--;	//Indices have been shifted after last deletion
		}

		if (deviceId.empty() || !SuggestedDriveLetter)
			return;

		String strHardwareId;
		const wchar_t wszGUID_DEVINTERFACE_VOLUME[] = L"{53f5630d-b6bf-11d0-94f2-00a0c91efb8b}";
		strHardwareId.Format(L"%s%s#%s", wszVCDPrefix, deviceId.c_str(), wszGUID_DEVINTERFACE_VOLUME);

		wchar_t wszKeyName[] = L"\\DosDevices\\X:";
		wszKeyName[12] = SuggestedDriveLetter & ~0x20;	//upcase()

		key[wszKeyName].SetValueEx(strHardwareId.c_str(), strHardwareId.length() * sizeof(wchar_t), REG_BINARY);
	}


	NTSTATUS ReportDeviceList(VirtualCDRecordHeader *pBuffer, const DynamicString &LetterOrImage, ULONG OutputLength, PULONG_PTR pBytesDone)
	{
		if (!pBuffer || (OutputLength < sizeof(VirtualCDRecordHeader)))
			return STATUS_BUFFER_TOO_SMALL;
		SynchronizedMapPointer pMap(*this);

		pBuffer->Version = VirtualCDRecordHeader::CurrentVersion;
		pBuffer->RecordCount = 0;

		size_t elCount = pMap->size();
		if (!elCount)
		{
			pBuffer->TotalSize = sizeof(VirtualCDRecordHeader);
			*pBytesDone = pBuffer->TotalSize;
			return STATUS_SUCCESS;
		}

		if (OutputLength < (sizeof(VirtualCDRecordHeader) + elCount * sizeof(VirtualCDRecord)))
		{
			pBuffer->TotalSize = sizeof(VirtualCDRecordHeader) + elCount * sizeof(VirtualCDRecord);
			*pBytesDone = sizeof(VirtualCDRecordHeader);
			return STATUS_SUCCESS;
		}

		size_t off = sizeof(VirtualCDRecordHeader) + elCount * sizeof(VirtualCDRecord);
		unsigned idx = 0;

		for (_MappingType::iterator it = pMap->begin(); it != pMap->end(); it++)
		{
			if (idx >= elCount)
				break;

			char driveLetter = 0;
			
			VirtualSCSIPDO *pSCSIPDO = static_cast<VirtualSCSIPDO *>(FindPDOByUniqueID(it->second));
			if (pSCSIPDO)
			{
				ManagedPointer<VirtualCDDevice> pDev = pSCSIPDO->GetRelatedSCSIDevice();
				if (pDev)
					driveLetter = pDev->GetLastReportedDriveLetter();
			}
			
			const DynamicString *pImageFile = &it->first;
			size_t stringSize = (pImageFile->length() + 1) * sizeof(wchar_t);

			if (!LetterOrImage.empty())
			{
				if (LetterOrImage.length() <= 2)
				{
					if ((LetterOrImage[0] | 0x20) != (driveLetter | 0x20))
						continue;
				}
				else
				{
					if (pImageFile->icompare(LetterOrImage))
						continue;
				}
			}
			
			if ((off + stringSize) <= OutputLength)
			{
				memcpy(((char *)pBuffer) + off, pImageFile->GetConstBuffer(), stringSize);
				pBuffer->Records[idx].ImageNameOffset = off;
				*pBytesDone = off + stringSize;
			}
			else
				pBuffer->Records[idx].ImageNameOffset = 0;
			
			off += stringSize;
			pBuffer->Records[idx].DriveLetter = driveLetter;
			idx++;
		}

		pBuffer->RecordCount = idx;
		pBuffer->TotalSize = off;
		*pBytesDone = off;
		return STATUS_SUCCESS;
	}


public:
	ImageFormats::ImageFormatDatabase m_ImageFormatDatabase;

	BazisVirtualCDBus() :
	  _GenericIndexedBus(L"BazisVirtualCDBus", false, FILE_DEVICE_SECURE_OPEN, FALSE, DO_BUFFERED_IO)
	{ 
	}

	virtual NTSTATUS OnDeviceControl(ULONG ControlCode, bool IsInternal, void *pInOutBuffer, ULONG InputLength, ULONG OutputLength, PULONG_PTR pBytesDone) override
	{
		wchar_t *pwszName = (wchar_t *)pInOutBuffer;
		bool bFound = false;
		//If NULL char is not found, our parameter is not a NULL-terminated string!
		for (unsigned i = 0; i < (InputLength / sizeof(wchar_t)); i++)
			if (!pwszName[i])
			{
				bFound = true;
				break;
			}
		if (!bFound && InputLength)
			return STATUS_INVALID_PARAMETER;
		switch (ControlCode)
		{
		case IOCTL_PLUGIN_DEVICE:
			{
				VirtualCDMountParams *pParams = (VirtualCDMountParams *)pInOutBuffer;
				if ((InputLength < sizeof(VirtualCDMountParams)) || !CheckStringLengthArray(pParams->wszFilePath))
					return STATUS_INVALID_PARAMETER;
				if (PDOExists(pParams->wszFilePath))
					return STATUS_OBJECT_NAME_COLLISION;

				ActionStatus st = MAKE_STATUS(NoMemory);
				ManagedPointer<AIFile> pFile = new OSFile(pParams->wszFilePath, FastFileFlags::OpenReadOnlyShareAll, &st);
				if (!pFile || !pFile->Valid())
					return st.ConvertToNTStatus();

				ManagedPointer<ImageFormats::AIParsedCDImage> pImage = m_ImageFormatDatabase.OpenCDImage(pFile, FilePath(pParams->wszFilePath));
				if (!pImage)
					return STATUS_BAD_FILE_TYPE;

				ManagedPointer<VirtualCDDevice> pDev = new VirtualCDDevice(pImage);
				if (!pDev)
					return STATUS_NO_MEMORY;

				const wchar_t *pDevName = wcsrchr(pParams->wszFilePath, '\\');
				if (!pDevName)
					pDevName = pParams->wszFilePath;
				else
					pDevName++;

				String devId = pDevName;
				for (int i = 0; i < devId.size(); i++)
					if (!wcschr(L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0124567890_", devId[i]))
						devId[i] = '_';

				CleanupOldLinksAndRegisterNewDeviceLink(devId, pParams->SuggestedDriveLetter);
				
				return InsertPDO(pParams->wszFilePath, new VirtualSCSIPDO(pDev, pDevName, devId.c_str(), true));
			}
		case IOCTL_PLUGOUT_DEVICE:
		case IOCTL_EJECT_DEVICE:
			if (!InputLength || !pwszName)
				return STATUS_INVALID_BUFFER_SIZE;
			else
				return (ControlCode == IOCTL_PLUGOUT_DEVICE) ? PlugoutPDO(pwszName) : EjectPDO(pwszName);
			break;
		case IOCTL_IS_DEVICE_PRESENT:
			if (OutputLength < 1)
				return STATUS_INVALID_BUFFER_SIZE;
			*((bool *)pInOutBuffer) = PDOExists(pwszName);
			*pBytesDone = 1;
			return STATUS_SUCCESS;
		case IOCTL_LIST_DEVICES:
			{
				DynamicString letterOrImageFilter = InputLength ? pwszName : _T("");
				if (OutputLength < sizeof(VirtualCDRecordHeader))
					return STATUS_INVALID_BUFFER_SIZE;
				return ReportDeviceList((VirtualCDRecordHeader *)pInOutBuffer, letterOrImageFilter, OutputLength, pBytesDone);
			}
		}

		return STATUS_INVALID_DEVICE_REQUEST;
	}

protected:
	virtual void ShutdownPDO(VirtualSCSIPDO *pPDO)
	{
		pPDO->ShutdownUnderlyingDevice();
	}
};

DDK::Driver *_stdcall CreateMainDriverInstance()
{
	return new _GenericPnpDriver<BazisVirtualCDBus, &GUID_BazisVirtualCDBus>();
}
